using System;
using Sandbox.Common.ObjectBuilders;
using Sandbox.ModAPI;
using Sandbox.ModAPI.Interfaces;
using Sandbox.ModAPI.Interfaces.Terminal;
using VRage.Game.Components;
using VRage.ModAPI;
using VRage.ObjectBuilders;
using VRageMath;

/*
 * Mod by rexxar
 * 
 * You are free to copy this code into your own mod so long as you do not remove this comment block.
 * 
 * This mod, combined with a simple extension class will allow programmable blocks to use
 * GetFreeDestination again. 
 */

namespace GFDExposer
{
    [MyEntityComponentDescriptor(typeof(MyObjectBuilder_RemoteControl))]
    public class GFDExposer : MyGameLogicComponent
    {
        private IMyRemoteControl _block;
        private static bool _init;
        private Vector3D _result;
        private Vector3D _destination;
        private float _checkRadius;
        private float _shipRadius;

        public override void Init(MyObjectBuilder_EntityBase objectBuilder)
        {
            _block = (IMyRemoteControl)Entity;
            Entity.NeedsUpdate |= MyEntityUpdateEnum.BEFORE_NEXT_FRAME;
        }

        public override void UpdateOnceBeforeFrame()
        {
            if (_init)
                return;

            _block = Entity as IMyRemoteControl;

            if (_block == null)
                return;

            _init = true;

            var destination = MyAPIGateway.TerminalControls.CreateProperty<Vector3D, IMyRemoteControl>("GFD_Exposer_Destination");
            destination.Setter = (b, v) =>
                                    {
                                        if (b.GameLogic.GetAs<GFDExposer>() != null)
                                            b.GameLogic.GetAs<GFDExposer>()._destination = v;
                                    };
            destination.Getter = b => b.GameLogic.GetAs<GFDExposer>()?._destination ?? Vector3D.Zero;
            MyAPIGateway.TerminalControls.AddControl<IMyRemoteControl>(destination);

            var checkRadius = MyAPIGateway.TerminalControls.CreateProperty<float, IMyRemoteControl>("GFD_Exposer_CheckRadius");
            checkRadius.Setter = (b, v) =>
                                    {
                                        if (b.GameLogic.GetAs<GFDExposer>() != null)
                                            b.GameLogic.GetAs<GFDExposer>()._checkRadius = v;
                                    };
            checkRadius.Getter = b => b.GameLogic.GetAs<GFDExposer>()?._checkRadius ?? 0;
            MyAPIGateway.TerminalControls.AddControl<IMyRemoteControl>(checkRadius);

            var shipRadius = MyAPIGateway.TerminalControls.CreateProperty<float, IMyRemoteControl>("GFD_Exposer_ShipRadius");
            shipRadius.Visible = b => false;
            shipRadius.Setter = (b, v) =>
                                   {
                                       if (b.GameLogic.GetAs<GFDExposer>() == null)
                                           return;
                                       b.GameLogic.GetAs<GFDExposer>()._shipRadius = v;
                                       DoGFD(b);
                                   };
            shipRadius.Getter = b => b.GameLogic.GetAs<GFDExposer>()?._shipRadius ?? 0;
            MyAPIGateway.TerminalControls.AddControl<IMyRemoteControl>(shipRadius);

            var result = MyAPIGateway.TerminalControls.CreateProperty<Vector3D, IMyRemoteControl>("GFD_Exposer_Result");
            result.Visible = b => false;
            result.Setter = (b, v) => { /* NOP */ };
            result.Getter = b => b.GameLogic.GetAs<GFDExposer>()?._result ?? Vector3D.Zero;
            MyAPIGateway.TerminalControls.AddControl<IMyRemoteControl>(result);
        }

        private void DoGFD(IMyTerminalBlock b)
        {
            var block = (IMyRemoteControl)b;
            Vector3D dest = b.GameLogic.GetAs<GFDExposer>()._destination;
            float check = b.GameLogic.GetAs<GFDExposer>()._checkRadius;
            float rad = b.GameLogic.GetAs<GFDExposer>()._shipRadius;

            Vector3D result = block.GetFreeDestination(dest, check, rad);

            b.GameLogic.GetAs<GFDExposer>()._result = result;
        }

        public override MyObjectBuilder_EntityBase GetObjectBuilder(bool copy = false)
        {
            return Entity.GetObjectBuilder(copy);
        }
    }
    
    /*
     * Add this extension class to your PB script, or simply copy the contents if you already have an extension class
     */

    public static class Extensions
    {
        public static Vector3D GetFreeDestination(this IMyRemoteControl block, Vector3D destination, float checkRadius, float shipRadius)
        {
            //no safety checks here. If the mod is not installed, the script should throw an exception
            //this is safer than returning 0,0,0 or some other garbage result
            block.SetValue("GFD_Exposer_Destination", destination);
            block.SetValue("GFD_Exposer_CheckRadius", checkRadius);
            block.SetValue("GFD_Exposer_ShipRadius", shipRadius);

            return block.GetValue<Vector3D>("GFD_Exposer_Result");
        }
    }
}
